void clear_screen(unsigned char color)
{
	memset(textBuf, 32, sizeof(textBuf));
	memset(attrBuf, color, sizeof(attrBuf));
}



int put_str(int x, int y, const char* str, unsigned char col)
{
	int xs = x;

	while (1)
	{
		char c = *str++;

		if (!c) break;

		switch (c)
		{
		case '~':
		{
			col = ((col << 4) & 0xf0) | ((col >> 4) & 0x0f);
			continue;
		}
		break;

		case '\n':
		{
			++y;
			x = xs;
		}
		break;

		default:
			if (x >= 0 && x < textScreenWidth) PUT_CHAR_ATTR(x, y, c, col);
			++x;
		}
	}

	return x - xs;
}



int count_str_wrap_lines(int wdt, int hgt, const char* str)
{
	int x = 0;
	int y = 0;

	unsigned int ptr = 0;

	while (ptr < strlen(str))
	{
		unsigned int ptr1 = ptr;
		int len = 0;

		while (ptr1 < strlen(str))
		{
			char c = str[ptr1];

			if (c == ' ' || c == '\n') break;

			++len;
			++ptr1;
		}

		if (x + len >= wdt)
		{
			x = 0;
			++y;
		}

		for (int i = 0; i < len; ++i)
		{
			++ptr;
			++x;
		}

		while (1)
		{
			char c = str[ptr];

			if (c == '\n')
			{
				x = 0;
				++y;
				++ptr;
				continue;
			}

			if (c == ' ')
			{
				++x;
				++ptr;
				continue;
			}

			break;
		}
	}

	return y;
}



void put_str_wrap(int xs, int ys, int wdt, int hgt, const char* str, unsigned char col, int line_off)
{
	unsigned int ptr = 0;

	int x = xs;
	int y = ys;

	while (ptr < strlen(str))
	{
		unsigned int ptr1 = ptr;
		int word_len = 0;

		while (ptr1 < strlen(str))
		{
			char c = str[ptr1];

			if (c == ' ' || c == '\n') break;

			++word_len;
			++ptr1;
		}

		if (x + word_len >= xs + wdt)
		{
			x = xs;
			if (line_off > 0) --line_off; else ++y;
		}

		if (line_off > 0)
		{
			x += word_len;
			ptr += word_len;
		}
		else
		{
			for (int i = 0; i < word_len; ++i)
			{
				PUT_CHAR_ATTR(x, y, str[ptr++], col);
				++x;
			}
		}

		while (1)
		{
			char c = str[ptr];

			if (c == '\n')
			{
				x = xs;
				if (line_off > 0) --line_off; else ++y;
				++ptr;
				continue;
			}

			if (c == ' ')
			{
				++x;
				++ptr;
				continue;
			}

			break;
		}

		if (hgt > 0 && (y - ys >= hgt)) break;
	}
}



int get_str_height(const char* str)
{
	int h = 1;

	while (1)
	{
		char c = *str++;

		if (!c) break;

		if (c == '\n') ++h;
	}

	return h;
}



void put_str_file(int x, int y, const char* str, unsigned char col)
{
	while (1)
	{
		unsigned char cb = *str++;

		char c = cb;

		if (x >= textScreenWidth || !c) break;

		if (c < 32) c = '?';

		PUT_CHAR_ATTR(x, y, c, col);

		if ((cb & 0xe0) == 0xc0) str += 1; else
			if ((cb & 0xf0) == 0xe0) str += 2; else
				if ((cb & 0xf8) == 0xf0) str += 3;

		++x;
	}
}



void put_num(int x, int y, int num, int digit)
{
	for (int i = 0; i < digit; ++i) PUT_CHAR(x + i, y, NUM2DECZ(num, digit - 1 - i));
}



void put_num_attr(int x, int y, int num, int digit, unsigned char color)
{
	for (int i = 0; i < digit; ++i) PUT_CHAR_ATTR(x + i, y, NUM2DECZ(num, digit - 1 - i), color);
}



inline unsigned char cursor_color(int row, int rowhl, int column)
{
	int x1, x2, y1, y2;
	unsigned char color;

	if (row == curRow && column == curColumn) return COL_CURSOR;

	if (!(rowhl%module.measure)) color = COL_PATTERN_BEAT; else color = COL_PATTERN_BACKGROUND;

	if (curSelectionXS >= 0)
	{
		if (curSelectionXS < curSelectionXE)
		{
			x1 = curSelectionXS;
			x2 = curSelectionXE;
		}
		else
		{
			x1 = curSelectionXE;
			x2 = curSelectionXS;
		}

		if (curSelectionYS < curSelectionYE)
		{
			y1 = curSelectionYS;
			y2 = curSelectionYE;
		}
		else
		{
			y1 = curSelectionYE;
			y2 = curSelectionYS;
		}

		if (column >= x1 && column <= x2 && row >= y1 && row <= y2)
		{
			if (!(row%module.measure)) color = COL_SELECTION_BEAT; else color = COL_SELECTION;

			if (row == curRow) color = COL_SELECTION_CURSOR;

			return color;
		}
	}

	if (row == curRow) color = COL_CURSOR_ROW;

	return color;
}



int get_pattern_top_row_2x(void)
{
	int row, hgt, rcnt, first_row, last_row;

	if (editStyle == EDIT_STYLE_LINEAR)
	{
		first_row = 0;
		last_row = MAX_ROWS;
	}
	else
	{
		first_row = curRow * 2;

		while (first_row >= 0)
		{
			if (module.marker[first_row / 2])
			{
				first_row += 1;
				break;
			}

			first_row -= 2;
		}

		if (first_row < 0) first_row = 0;

		last_row = curRow;

		while (last_row < MAX_ROWS)
		{
			++last_row;

			if (module.marker[last_row]) break;
		}

		if (last_row > MAX_ROWS - 1) last_row = MAX_ROWS - 1;
	}

	hgt = PATTERN_HEIGHT;

	if (!hgt) hgt = 1;

	rcnt = PATTERN_HEIGHT / 2;
	row = curRow * 2;

	while (rcnt > 0)
	{
		row -= 2;

		if (row < first_row) break;

		if (module.marker[row / 2]/*||row/2==songList[SongCur].length*/) --rcnt;

		--rcnt;
	}

	if (rcnt < 0) ++row;

	if (module.marker[curRow]) row += 2;

	if (row < first_row) row = first_row;

	if (row / 2 + PATTERN_HEIGHT + 1 >= last_row) row = (last_row - PATTERN_HEIGHT + 1) * 2;

	if (row < first_row) row = first_row;

	return row;
}



void draw_marker_name(int x, int y, int row, int color, bool ralign)
{
	int len = strlen(module.markerName[row]);

	if (!len) return;

	if (ralign) x -= len;

	for (int i = 0; i < len; ++i) PUT_CHAR_ATTR(x + i, y, module.markerName[row][i], color);
}



bool module_is_row_empty(int track, int row, bool marker);

int draw_column_position(int x, int y, int track, int row, int col, int ct, int color)
{
	switch (ct)
	{
	case COLUMN_TYPE_EMPTY:
	{
		PUT_CHAR_ATTR(x + 0, y, ' ', color);

		++x;
	}
	break;

	case COLUMN_TYPE_LETTER:
	{
		unsigned char val = module.rows[track][row].column[col];
		PUT_CHAR_ATTR(x + 0, y, val ? val : '.', color);

		++x;
	}
	break;

	case COLUMN_TYPE_NOTE:
	{
		unsigned char note = module.rows[track][row].column[col] & 127;

		PUT_CHAR_ATTR(x + 0, y, noteNames[note][0], color);
		PUT_CHAR_ATTR(x + 1, y, noteNames[note][1], color);
		PUT_CHAR_ATTR(x + 2, y, noteNames[note][2], color);

		x += 3;
	}
	break;

	default:
	{
		unsigned char val = module.rows[track][row].column[col];

		for (int i = 0; i < ct; ++i)
		{
			PUT_CHAR_ATTR(x + i, y, NUM2DEC(val, ct - 1 - i), color);
		}

		x += module.columnType[col];
	}
	}

	return x;
}



void draw_pattern(void)
{
	if (!module.columns) return;

	int row2x_orig = get_pattern_top_row_2x();

	int rowhl_orig = 0;

	for (int row = 0; row < row2x_orig / 2; ++row)
	{
		if (module.marker[row]) rowhl_orig = 0;

		++rowhl_orig;
	}

	int y = PATTERN_YPOS;

	int row2x = row2x_orig;
	int rowhl = rowhl_orig;

	while (y < (PATTERN_YPOS + PATTERN_HEIGHT))
	{
		int x = PATTERN_XPOS + 1;

		int row = row2x / 2;

		if (row >= MAX_ROWS) break;

		if (!(row2x & 1))
		{
			if (editStyle == EDIT_STYLE_CHOPPED)
			{
				if (module.marker[row]) break;
			}
			else
			{
				if (row && (module.marker[row] || (module.loopEnd < MAX_ROWS - 1 && row == module.loopEnd)))
				{
					++y;
				}
			}
		}
		else
		{
			if (module.marker[row]) rowhl = 0;

			unsigned char color = cursor_color(row, rowhl, -1);

			int row_display = row;

			if (editStyle != EDIT_STYLE_LINEAR)
			{
				if (rowhl) row_display = rowhl;
			}

			if (!module_is_row_empty(activeTrack ^ 1, row_display, false)) PUT_CHAR(0, y, !activeTrack ? '>' : '<');

			PUT_CHAR_ATTR(x + 0, y, (module.loopSupport&&row == module.loopStart) ? '>' : ' ', color);
			PUT_CHAR_ATTR(x + 1, y, NUM2DECZ(row_display, 3), color);
			PUT_CHAR_ATTR(x + 2, y, NUM2DECZ(row_display, 2), color);
			PUT_CHAR_ATTR(x + 3, y, NUM2DECZ(row_display, 1), color);
			PUT_CHAR_ATTR(x + 4, y, NUM2DECZ(row_display, 0), color);
			PUT_CHAR_ATTR(x + 5, y, (module.loopSupport&&row == module.loopEnd) ? '<' : ' ', color);

			x += 6;

			for (int col = 0; col < module.columns; ++col)
			{
				int ct = module.columnType[col];

				int color = cursor_color(row, rowhl, col);

				x = draw_column_position(x, y, activeTrack, row, col, ct, color);
			}

			color = cursor_color(row, rowhl, -1);

			PUT_CHAR_ATTR(x, y, ' ', color);

			++y;
			++rowhl;
		}

		++row2x;
	}
}



void draw_pattern_markers(void)
{
	if (!module.columns) return;

	int row2x_orig = get_pattern_top_row_2x();

	int rowhl_orig = 0;

	for (int row = 0; row < row2x_orig / 2; ++row)
	{
		if (module.marker[row]) rowhl_orig = 0;

		++rowhl_orig;
	}

	int y = PATTERN_YPOS;

	int row2x = row2x_orig;

	while (y < (PATTERN_YPOS + PATTERN_HEIGHT))
	{
		int x = PATTERN_XPOS + strlen(engineHeader) + 3;

		int row = row2x / 2;

		if (row >= MAX_ROWS) break;

		if (!(row2x & 1))
		{
			if (editStyle == EDIT_STYLE_CHOPPED)
			{
				if (module.marker[row]) break;
			}
			else
			{
				if (row && (module.marker[row] || (module.loopEnd < MAX_ROWS - 1 && row == module.loopEnd)))
				{
					++y;
				}
			}
		}
		else
		{
			draw_marker_name(x, y, row, COL_MARKER_LIST, false);

			++y;
		}

		++row2x;
	}
}



void draw_ghost_pattern(void)
{
	if (!module.columns) return;

	int row2x_orig = get_pattern_top_row_2x();

	int rowhl_orig = 0;

	for (int row = 0; row < row2x_orig / 2; ++row)
	{
		if (module.marker[row]) rowhl_orig = 0;

		++rowhl_orig;
	}

	int y = PATTERN_YPOS;

	int row2x = row2x_orig;

	while (y < (PATTERN_YPOS + PATTERN_HEIGHT))
	{
		int x = PATTERN_XPOS + strlen(engineHeader) + 3;

		int row = row2x / 2;

		if (row >= MAX_ROWS) break;

		if (!(row2x & 1))
		{
			if (editStyle == EDIT_STYLE_CHOPPED)
			{
				if (module.marker[row]) break;
			}
			else
			{
				if (row && (module.marker[row] || (module.loopEnd < MAX_ROWS - 1 && row == module.loopEnd)))
				{
					++y;
				}
			}
		}
		else
		{
			for (int col = 0; col < module.columns; ++col)
			{
				int ct = module.columnType[col];

				int color = COL_MARKER_LIST;

				x = draw_column_position(x, y, activeTrack ^ 1, row, col, ct, color);
			}

			++y;
		}

		++row2x;
	}
}



void update_marker_list(void)
{
	markerList.row[0] = 0;
	markerList.length = 1;
	markerList.cur = 0;

	if (module.loopSupport)
	{
		markerList.loopStart = 0;
		markerList.loopEnd = 0;
	}
	else
	{
		markerList.loopStart = -1;
		markerList.loopEnd = -1;
	}

	for (int row = 1; row < MAX_ROWS; ++row)
	{
		if (module.marker[row])
		{
			markerList.row[markerList.length] = row;
			if (module.loopSupport)
			{
				if (module.loopStart >= row) markerList.loopStart = markerList.length;
				if (module.loopEnd >= row) markerList.loopEnd = markerList.length;
			}
			++markerList.length;
		}

		if (row == curRow)
		{
			markerList.cur = markerList.length - 1;
		}
	}
}



void draw_marker_list(void)
{
	update_marker_list();

	int x = textScreenWidth - 5;
	int y = PATTERN_YPOS;

	int off = 0;

	if (markerList.length > PATTERN_HEIGHT)
	{
		off = markerList.cur - PATTERN_HEIGHT / 2;

		if (off < 0) off = 0;

		if (off >= markerList.length - PATTERN_HEIGHT) off = markerList.length - PATTERN_HEIGHT;
	}

	for (int i = 0; i < PATTERN_HEIGHT; ++i)
	{
		int row = markerList.row[off];

		int color = 0;

		if (markerList.cur == off) color = COL_PROGRAM_BACKGROUND; else color = COL_MARKER_LIST;

		if (off == markerList.loopStart) PUT_CHAR_ATTR(x - 1, y, '>', color);
		PUT_CHAR_ATTR(x + 0, y, NUM2DECZ(row, 3), color);
		PUT_CHAR_ATTR(x + 1, y, NUM2DECZ(row, 2), color);
		PUT_CHAR_ATTR(x + 2, y, NUM2DECZ(row, 1), color);
		PUT_CHAR_ATTR(x + 3, y, NUM2DECZ(row, 0), color);
		if (off == markerList.loopEnd) PUT_CHAR_ATTR(x + 4, y, '<', color);

		draw_marker_name(x - 1, y, markerList.row[off], color, true);

		++y;
		++off;

		if (off >= markerList.length) break;
	}
}



int draw_menu_item(int x, const char *str)
{
	x += put_str(x, textScreenHeight - 1, str, COL_HEADER);

	return x;
}



void draw_settings(void)
{
	for (int i = 0; i < textScreenWidth; ++i)
	{
		PUT_CHAR_ATTR(i, 0, ' ', COL_HEADER);
		PUT_CHAR_ATTR(i, textScreenHeight - 1, ' ', COL_HEADER);
	}

	put_str(0, textScreenHeight - 1, " Oct0 Spd00 Stp00", COL_HEADER);

	int x = 20;

	if (!altMenu)
	{
		x = draw_menu_item(x, "~F1~Help");
		x = draw_menu_item(x, "~F2~Save");
		x = draw_menu_item(x, "~F3~Load");

		if (module.instruments) x = draw_menu_item(x, "~F4~Instr");

		x = draw_menu_item(x, "~F5~Play");
		x = draw_menu_item(x, "~F6~Info");
		x = draw_menu_item(x, "~F7~Engine");
		x = draw_menu_item(x, "~F8~Export");
	}
	else
	{
		x = draw_menu_item(x, (editStyle != EDIT_STYLE_CHOPPED) ? "~F1~-Chop" : "~F1~+Chop");

		x = draw_menu_item(x, !quietMode ? "~F2~-Quiet" : "~F2~+Quiet");

		x = draw_menu_item(x, "~F3~Zap");

		if (module.speed_column >= 0)
		{
			x = draw_menu_item(x, "~F4~RowOpt");
		}

		x = draw_menu_item(x, !dualMode ? "~F5~-Dual" : "~F5~+Dual");
		x = draw_menu_item(x, !activeTrack ? "~F6~Main" : "~F6~Side");

		x = draw_menu_item(x, "~F8~ExpWAV");
		x = draw_menu_item(x, "~F10~Quit");
	}

	put_num(4, textScreenHeight - 1, curOctave, 1);
	put_num(9, textScreenHeight - 1, module.speed, 2);
	put_num(15, textScreenHeight - 1, curAutoStep, 2);

	if (!module.columns)
	{
		put_str(1, 0, "Loading...", COL_HEADER);
		return;
	}

	put_str(1, 0, engineHeader, COL_HEADER);

	PUT_CHAR(0, 0, !activeTrack ? 'M' : 'S');

	x = PATTERN_XPOS + 7;

	for (int i = 0; i < module.columns; ++i)
	{
		int len = module.columnType[i];

		switch (len)
		{
		case COLUMN_TYPE_EMPTY: len = 1; break;
		case COLUMN_TYPE_LETTER: len = 1; break;
		case COLUMN_TYPE_NOTE:  len = 3; break;
		}

		if (module.columnChannel[i] >= 0)
		{
			if (channelMute[activeTrack][module.columnChannel[i]])
			{
				for (int j = 0; j < len; ++j) PUT_ATTR(x + j, 0, COL_HEADER_MUTE);
			}
		}

		x += len;
	}
}



void set_compile_error(const char* error_str)
{
	strncpy(engineErrorMessage, error_str, sizeof(engineErrorMessage));

	engineErrorMessageActive = true;
}



void clear_compile_error(void)
{
	strncpy(engineErrorMessage, "", sizeof(engineErrorMessage));
	engineErrorMessageActive = false;
}



void draw_compile_error(void)
{
	if (!engineErrorMessageActive) return;

	for (int i = 0; i < textScreenWidth; ++i) PUT_CHAR_ATTR(i, 0, ' ', COL_ERROR_BACKGROUND);

	put_str(5, 0, engineErrorMessage, COL_ERROR_BACKGROUND);

	put_str(1, 0, "~ESC~", COL_ERROR_BACKGROUND);

	int lines = 0;

	const char* log = log_recent();

	if (strlen(log) > 0) lines += 2;

	for (unsigned int i = 0; i < strlen(log); ++i)
	{
		if (log[i] == '\n') ++lines;
	}

	if (lines > MAX_COMPILE_ERROR_LINES) lines = MAX_COMPILE_ERROR_LINES;

	for (int i = 0; i < lines; ++i)
	{
		for (int j = 0; j < textScreenWidth; ++j) PUT_CHAR_ATTR(j, 1 + i, ' ', COL_PATTERN_BACKGROUND);
	}

	put_str_wrap(1, 2, textScreenWidth - 2, lines - 2, log, COL_PATTERN_BACKGROUND, 0);
}