#include "playlayer.h"
#include <stdlib.h>


void critical_error(const char *format, ...);
void log_add(const char *format, ...);



void PlayLayer::PlayLayerHandleGMEError(gme_err_t error, const char *name)
{
	if (error) critical_error("[GME] %s %s\n", name, error);
}



PlayLayer::PlayLayer()
{
	playing = false;
}



PlayLayer::~PlayLayer()
{
}



void PlayLayer::Play(const unsigned char * data, size_t data_size, int sample_rate)
{
	backend = PLL_BACKEND_GME;

	//if (PLL_Z80CheckSignature(data, data_size)) backend = PLL_BACKEND_Z80;

	if (playing) Stop();

	switch (backend)
	{
	case PLL_BACKEND_GME:
		PlayLayerHandleGMEError(gme_open_data(data, data_size, &gme_emu, sample_rate), "gme_open_data");
		gme_ignore_silence(gme_emu, 1);
		PlayLayerHandleGMEError(gme_start_track(gme_emu, 0), "gme_start_track");
		playing = true;
		break;

	case PLL_BACKEND_Z80:
		break;

	default:
		log_add("Warning: attempted to play an unknown audio container format\n");
	}
}



void PlayLayer::Stop()
{
	switch (backend)
	{
	case PLL_BACKEND_GME:
		if (gme_emu)
		{
			gme_delete(gme_emu);
			gme_emu = NULL;
		}
		break;

	case PLL_BACKEND_Z80:
		break;
	}

	playing = false;
}



void PlayLayer::Render(short int* buf, int samples)
{
	switch (backend)
	{
	case PLL_BACKEND_GME:
		PlayLayerHandleGMEError(gme_play(gme_emu, samples, buf), "gme_play");
		break;

	case PLL_BACKEND_Z80:
		break;
	}
}
