char blockToCopyName[MAX_MARKER_NAME + 1]; 
bool blockToCopyShift;


void set_undo(void)
{
	memcpy(&moduleUndo,&module,sizeof(module));

	undoAvailable=true;
}



void undo(void)
{
	if(undoAvailable)
	{
		unsigned char* src = (unsigned char*)&module;
		unsigned char* dst = (unsigned char*)&moduleUndo;

		for (int i = 0; i < sizeof(module); ++i)
		{
			unsigned char temp = *src;
			*src++ = *dst;
			*dst++ = temp;
		}
	}
}



void move_row_cursor(int off)
{
	curRow += off;

	if (curRow < 0) curRow = 0;
	if (curRow >= MAX_ROWS) curRow = MAX_ROWS - 1;

	curAutoStepCount = 0;
	curResetValue = true;
}



void move_row_cursor_prev_marker(void)
{
	module.marker[0] |= 2;
	module.marker[module.loopStart] |= 2;
	module.marker[module.loopEnd] |= 2;

	for (int i = curRow - 1; i >= 0; --i)
	{
		if(module.marker[i])
		{
			curRow=i;
			break;
		}
	}

	module.marker[0] &= ~2;
	module.marker[module.loopStart] &= ~2;
	module.marker[module.loopEnd] &= ~2;
}



void move_row_cursor_next_marker(void)
{
	module.marker[MAX_ROWS - 1] |= 2;
	module.marker[module.loopStart] |= 2;
	module.marker[module.loopEnd] |= 2;

	for (int i = curRow + 1; i < MAX_ROWS; ++i)
	{
		if(module.marker[i])
		{
			curRow=i;
			break;
		}
	}

	module.marker[MAX_ROWS-1]&=~2;
	module.marker[module.loopStart]&=~2;
	module.marker[module.loopEnd  ]&=~2;
}



int get_song_first_row(void)
{
	if (editStyle == EDIT_STYLE_CHOPPED)
	{
		int first_row = curRow;

		while (first_row > 0)
		{
			if (module.marker[first_row]) break;

			--first_row;
		}

		if (first_row < 0) first_row = 0;

		return first_row;
	}
	else
	{
		return 0;
	}
}



int get_song_last_row(void)
{
	if (editStyle == EDIT_STYLE_CHOPPED)
	{
		int last_row = curRow;

		while (last_row < MAX_ROWS)
		{
			++last_row;

			if (module.marker[last_row]) break;
		}

		if (last_row > MAX_ROWS - 1) last_row = MAX_ROWS - 1;

		return last_row;
	}
	else
	{
		return MAX_ROWS;
	}
}



void delete_pos_column(int row,int column)
{
	if (row < get_song_first_row() || row >= get_song_last_row()) return;

	if(column<0)
	{
		for (int i = row - 1; i < get_song_last_row() - 1; ++i)
		{
			if (i >= 0)
			{
				module.marker[i] = module.marker[i + 1];
				strncpy(module.markerName[i], module.markerName[i + 1], sizeof(module.markerName[i]));
			}
		}

		module.marker[get_song_last_row() - 1] = 0;
		module.markerName[get_song_last_row() - 1][0] = 0;
	}
	else
	{
		for (int i = row - 1; i < get_song_last_row() - 1; ++i)
		{
			if (i >= 0)
			{
				module.rows[activeTrack][i].column[column] = module.rows[activeTrack][i + 1].column[column];
			}
		}

		module.rows[activeTrack][get_song_last_row() - 1].column[column] = 0;
	}
}



void insert_pos_column(int row,int column)
{
	if (row < get_song_first_row() || row >= get_song_last_row()) return;

	if(column<0)
	{
		for (int i = get_song_last_row() - 1; i > row; --i)
		{
			module.marker[i] = module.marker[i - 1];
			strncpy(module.markerName[i], module.markerName[i - 1], sizeof(module.markerName[i]));
		}

		module.marker[row] = 0;
		module.markerName[row][0] = 0;
	}
	else
	{
		for (int i = get_song_last_row() - 1; i > row; --i) module.rows[activeTrack][i].column[column] = module.rows[activeTrack][i - 1].column[column];

		module.rows[activeTrack][row].column[column]=0;
	}
}



void adjust_loop_points(int cur_row,int step)
{
	if (module.loopStart >= cur_row)
	{
		module.loopStart += step;

		if (module.loopStart < 0) module.loopStart = 0;
		if (module.loopEnd < 0) module.loopEnd = 0;
	}

	if (module.loopEnd >= cur_row)
	{
		module.loopEnd += step;

		if (module.loopStart > MAX_ROWS) module.loopStart = MAX_ROWS;
		if (module.loopEnd > MAX_ROWS) module.loopEnd = MAX_ROWS;
	}
}



void insert_row_with_shift(int row)
{
	for (int col = -1; col < module.columns; ++col) insert_pos_column(row, col);	//col -1 for markers

	adjust_loop_points(row, 1);
}



void delete_row_with_shift(int row)
{
	for (int col = -1; col < module.columns; ++col) delete_pos_column(row, col);

	adjust_loop_points(row, -1);
}



void update_selection(void)
{
	if(curShiftHold)
	{
		curSelectionXE=curColumn;
		curSelectionYE=curRow;

		if(curSelectionXS<curSelectionXE)
		{
			curSelectionX=curSelectionXS;
			curSelectionWidth=curSelectionXE-curSelectionXS+1;
		}
		else
		{
			curSelectionX=curSelectionXE;
			curSelectionWidth=curSelectionXS-curSelectionXE+1;
		}

		if(curSelectionYS<curSelectionYE)
		{
			curSelectionY=curSelectionYS;
			curSelectionHeight=curSelectionYE-curSelectionYS+1;
		}
		else
		{
			curSelectionY=curSelectionYE;
			curSelectionHeight=curSelectionYS-curSelectionYE+1;
		}
	}

	if(!curShiftHold||!curSelectionActive)
	{
		curSelectionActive=false;
		curSelectionXS=-1;
		curSelectionYS=-1;
		curSelectionXE=-1;
		curSelectionYE=-1;
		curSelectionWidth=0;
		curSelectionHeight=0;
	}

	curAutoStepCount=0;
	curResetValue=true;
}



void set_selection(int x,int y,int wdt,int hgt)
{
	curSelectionX = x;
	curSelectionY = y;
	curSelectionXS=x;
	curSelectionYS=y;
	curSelectionXE=x+wdt-1;
	curSelectionYE=y+hgt-1;
	curSelectionWidth=wdt;
	curSelectionHeight=hgt;

	curSelectionActive = true;
}



void select_channel(int col)
{
	int i,x,y,w,h,newh,channel,cmin,cmax;

	cmin = curRow;
	cmax = curRow;

	while (cmin > 0)
	{
		if (module.marker[cmin] || (module.loopEnd < (MAX_ROWS - 1) && cmin == module.loopEnd)) break;

		--cmin;
	}

	while (cmax < MAX_ROWS)
	{
		++cmax; 
		
		if (module.marker[cmax] || (module.loopEnd < (MAX_ROWS - 1) && cmax == module.loopEnd)) break;
	}

	if (cmin == 0 && cmax == MAX_ROWS&&curSelectionHeight == MAX_ROWS)
	{
		cmin = 0;
		cmax = 0;
	}

	newh = cmax - cmin;

	if (!curSelectionActive || (curSelectionY != cmin&&curSelectionHeight != newh) || curSelectionHeight==MAX_ROWS)
	{		
		y = cmin;
		h = newh;
	}
	else
	{
		y = 0;
		h = MAX_ROWS;
	}

	if (col >= 0)
	{
		channel = module.columnChannel[col];

		if (channel < 0)
		{
			x = curColumn;
			w = 1;
		}
		else
		{
			cmin = module.columns;
			cmax = 0;

			for (i = 0; i < module.columns; ++i)
			{
				if (module.columnChannel[i] == channel)
				{
					if (i < cmin) cmin = i;
					if (i > cmax) cmax = i;
				}
			}

			x = cmin;
			w = cmax - cmin + 1;
		}
	}
	else
	{
		x = 0;
		w = module.columns;
	}

	set_selection(x, y, w, h);
}



void selection_copy_cut(bool copy,bool cut,bool shift)
{
	if(cut) set_undo();

	for (int col = 0; col < curSelectionWidth; ++col)
	{
		copyPasteType[col] = module.columnType[curSelectionX + col];

		if (copy)
		{
			for (int row = 0; row < curSelectionHeight; ++row)
			{
				copyPasteData[row].column[col] = module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col];

				if (!col)
				{
					copyPasteMarker[row] = module.marker[curSelectionY + row];
					if ((curSelectionY + row) == 0) copyPasteMarker[row] = 1;

					memcpy(copyPasteMarkerName[row], module.markerName[curSelectionY + row], sizeof(module.markerName[0]));
				}
			}
		}

		if(cut)
		{
			for (int row = 0; row < curSelectionHeight; ++row) module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col] = 0;

			if(shift)
			{
				for (int row = curSelectionY; row < get_song_last_row(); ++row) module.rows[activeTrack][row].column[curSelectionX + col] = module.rows[activeTrack][row + curSelectionHeight].column[curSelectionX + col];

				for (int row = get_song_last_row() - curSelectionHeight; row < get_song_last_row(); ++row) module.rows[activeTrack][row].column[curSelectionX + col] = 0;
			}
		}
	}

	if(copy)
	{
		copyPasteWidth = curSelectionWidth;
		copyPasteHeight = curSelectionHeight;
	}

	if (cut&&shift)
	{
		if (curSelectionWidth == module.columns)
		{
			for (int row = curSelectionY; row < get_song_last_row() - curSelectionHeight; ++row)
			{
				module.marker[row] = module.marker[row + curSelectionHeight];
				strncpy(module.markerName[row], module.markerName[row + curSelectionHeight], sizeof(module.markerName[row]));
			}

			for (int row = get_song_last_row() - curSelectionHeight; row < get_song_last_row(); ++row)
			{
				module.marker[row] = 0;
				strncpy(module.markerName[row], "", sizeof(module.markerName[row]));
			}
			
			adjust_loop_points(curRow, -curSelectionHeight);	 
		}

		curRow = curSelectionY;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_paste(bool shift)
{
	set_undo();

	if (shift)
	{
		if (copyPasteWidth == module.columns)
		{
			for (int row = get_song_last_row() - 1 - copyPasteHeight; row >= curRow; --row)
			{
				module.marker[row + copyPasteHeight] = module.marker[row];
				memcpy(module.markerName[row + copyPasteHeight], module.markerName[row], sizeof(module.markerName[0]));
			}

			adjust_loop_points(curRow, copyPasteHeight);
		}
	}

	for (int col = 0; col < copyPasteWidth; ++col)
	{
		if(shift)
		{
			for (int row = get_song_last_row() - 1 - copyPasteHeight; row >= curRow; --row)
			{
				module.rows[activeTrack][row + copyPasteHeight].column[curColumn + col] = module.rows[activeTrack][row].column[curColumn + col];
			}
		}

		int type = copyPasteType[col];

		if (module.columnType[curColumn + col] == type)
		{
			for (int row = 0; row < copyPasteHeight; ++row)
			{
				int rr = curRow + row;

				if (rr < get_song_first_row() || rr >= get_song_last_row()) continue;

				int value = copyPasteData[row].column[col];

				switch(type)
				{
				case COLUMN_TYPE_NOTE:
					{
						if(value>1)
						{
							if (value < module.lowestNote)  value = module.lowestNote;
							if (value > module.highestNote) value = module.highestNote;
						}
					}
					break;

				case COLUMN_TYPE_LETTER:
					{
					//do nothing special
					}
					break;

				default:
					{
					if (value > module.columnMaxValue[curColumn + col]) value = module.columnMaxValue[curColumn + col];
					}
				}

				module.rows[activeTrack][rr].column[curColumn + col] = value;

				if (!col&&copyPasteWidth == module.columns)
				{
					module.marker[rr] = copyPasteMarker[row];

					memcpy(module.markerName[rr], copyPasteMarkerName[row], sizeof(module.markerName[0]));
				}
			}
		}
	}

	curSelectionActive = false;

	update_selection();
}



void selection_expand(void)
{
	if (curSelectionHeight < 2) return;

	set_undo();

	int row = curSelectionY + 1;

	for (int i = 0; i < curSelectionHeight; ++i)
	{
		for (int col = 0; col < curSelectionWidth; ++col) insert_pos_column(row, curSelectionX + col);

		if (curSelectionWidth == module.columns) insert_pos_column(row, -1);

		row += 2;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_shrink(void)
{
	if(curSelectionHeight<2) return;

	set_undo();

	int row = curSelectionY + 2;

	for (int i = 0; i < (curSelectionHeight >> 1); ++i)
	{
		for (int col = 0; col < curSelectionWidth; ++col) delete_pos_column(row, curSelectionX + col);

		if (curSelectionWidth == module.columns) delete_pos_column(row, -1);

		++row;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_transpose(int trans)
{
	set_undo();

	for (int row = 0; row < curSelectionHeight; ++row)
	{
		for (int col = 0; col < curSelectionWidth; ++col)
		{
			if (module.columnType[curSelectionX + col] == COLUMN_TYPE_NOTE)
			{
				int note = module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col];

				if (note > 1)
				{
					note += trans;

					if (note < module.lowestNote)  note = module.lowestNote;
					if (note > module.highestNote) note = module.highestNote;

					module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col] = note;
				}
			}
		}
	}
}



void selection_replicate(void)
{
	set_undo();

	for (int col = 0; col < curSelectionWidth; ++col)
	{
		for (int row = get_song_last_row() - 1; row >= curSelectionY + curSelectionHeight; --row)
		{
			module.rows[activeTrack][row].column[curSelectionX + col] = module.rows[activeTrack][row - curSelectionHeight].column[curSelectionX + col];
		}
	}

	if (curSelectionWidth == module.columns)
	{
		for (int row = get_song_last_row() - 1; row >= curSelectionY + curSelectionHeight; --row)
		{
			module.marker[row] = module.marker[row - curSelectionHeight];
			strncpy(module.markerName[row], module.markerName[row - curSelectionHeight], sizeof(module.markerName[row]));
		}
	}

	curSelectionY += curSelectionHeight;
	curSelectionYS += curSelectionHeight;
	curSelectionYE += curSelectionHeight;

	if (curSelectionY >= get_song_last_row()) curSelectionY = get_song_last_row() - 1;
	if (curSelectionYS >= get_song_last_row()) curSelectionYS = get_song_last_row() - 1;
	if (curSelectionYE > get_song_last_row()) curSelectionYE = get_song_last_row();

	curRow += curSelectionHeight;

	if (curRow >= get_song_last_row()) curRow = get_song_last_row() - 1;
}



void change_channel(int dir)
{
	int col = module.columnChannel[curColumn];

	if(dir<0)
	{
		while (col == module.columnChannel[curColumn] || module.columnType[curColumn] == COLUMN_TYPE_EMPTY)
		{
			--curColumn;

			if(curColumn<0) curColumn=module.columns-1;
		}
	}
	else
	{
		while (col == module.columnChannel[curColumn] || module.columnType[curColumn] == COLUMN_TYPE_EMPTY)
		{
			++curColumn;

			if(curColumn>=module.columns) curColumn=0;
		}
	}
}



int get_cursor_row_y(void)
{
	int row2x = get_pattern_top_row_2x();
	int y = PATTERN_YPOS;

	while (y < (PATTERN_YPOS + PATTERN_HEIGHT))
	{
		int row = row2x / 2;

		if (row >= MAX_ROWS) break;

		if (!(row2x & 1))
		{
			if (row && (module.marker[row] || (module.loopEnd < MAX_ROWS - 1 && row == module.loopEnd)))
			{
				++y;
			}
		}
		else
		{
			if (row == curRow) break;

			++y;
		}

		++row2x;
	}

	return y;
}



int get_pattern_width(void)
{
	int w = 0;

	for (int col = 0; col < module.columns; ++col)
	{
		int ct = module.columnType[col];

		switch (ct)
		{
		case COLUMN_TYPE_EMPTY: ++w; break;
		case COLUMN_TYPE_LETTER: ++w; break;
		case COLUMN_TYPE_NOTE: w += 3; break;
		default: w += module.columnType[col];
		}
	}

	return w;
}



void unmute_all(void)
{
	for (int col = 0; col < MAX_COLUMNS; ++col) channelMute[activeTrack][col] = false;
}



void mute_all_but_current(void)
{
	for (int col = 0; col < MAX_COLUMNS; ++col)
	{
		channelMute[activeTrack][col] = (col != module.columnChannel[curColumn]) ? true : false;
	}
}



void selection_borrow_tracks(bool over)
{
	set_undo();

	for (int col = 0; col < curSelectionWidth; ++col)
	{
		if (module.columnType[curSelectionX + col] == COLUMN_TYPE_NOTE)
		{
			for (int row = 0; row < curSelectionHeight; ++row)
			{
				int new_note = module.rows[activeTrack ^ 1][curSelectionY + row].column[curSelectionX + col];

				if (new_note > 0 && (module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col] == 0 || over))
				{
					module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col] = new_note;
				}
			}
		}
	}
}



void selection_exchange_tracks(void)
{
	set_undo();

	if (curSelectionWidth > 0 && curSelectionHeight > 0)
	{
		for (int col = 0; col < curSelectionWidth; ++col)
		{
			for (int row = 0; row < curSelectionHeight; ++row)
			{
				int c0 = module.rows[0][curSelectionY + row].column[curSelectionX + col];
				int c1 = module.rows[1][curSelectionY + row].column[curSelectionX + col];
				module.rows[0][curSelectionY + row].column[curSelectionX + col] = c1;
				module.rows[1][curSelectionY + row].column[curSelectionX + col] = c0;
			}
		}
	}
	else
	{
		int c0 = module.rows[0][curRow].column[curColumn];
		int c1 = module.rows[1][curRow].column[curColumn];
		module.rows[0][curRow].column[curColumn] = c1;
		module.rows[1][curRow].column[curColumn] = c0;
	}
}



void selection_unwind_notes(void)
{
	set_undo();

	for (int col = 0; col < curSelectionWidth; ++col)
	{
		if(module.columnType[curSelectionX + col]==COLUMN_TYPE_NOTE)
		{
			int note = 0;

			for (int row = 0; row < curSelectionHeight; ++row)
			{
				int new_note = module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col];

				if (new_note > 0) note = new_note; else module.rows[activeTrack][curSelectionY + row].column[curSelectionX + col] = note;
			}
		}
	}
}



int block_find_beginning(int row)
{
	while (1)
	{
		if (module.marker[row]) break;
		if (!row) break;

		--row;
	}

	return row;
}



int block_find_by_name(const char* name)
{
	for (int row = 0; row < MAX_ROWS; ++row)
	{
		if (module.marker[row] || !row)
		{
			if (strcmp(module.markerName[row], name) == 0)
			{
				return row;
			}
		}
	}

	return -1;
}



int block_get_length(int src_row)
{
	for (int row = src_row + 1; row < MAX_ROWS; ++row)
	{
		if (module.marker[row])
		{
			return row - src_row;
		}
	}

	return 0;
}



void block_set_name(void)
{
	int row = block_find_beginning(curRow);

	altMenu = false;

	set_input_string(PATTERN_XPOS + 9 + get_pattern_width(), get_cursor_row_y(), MAX_MARKER_NAME + 1, MODE_EDIT_SONG, true, module.markerName[row], NULL);
}



void block_copy_by_name_cb(void)
{
	int src_row = block_find_by_name(blockToCopyName);

	if (src_row >= 0)
	{
		int src_len = block_get_length(src_row);

		if (src_len > 0)
		{
			set_undo();

			rowStruct* temp = (rowStruct*)malloc(src_len * sizeof(rowStruct));
			char new_marker[MAX_MARKER_NAME + 1];

			strncpy(new_marker, "@", sizeof(new_marker));
			strncat(new_marker, module.markerName[src_row], sizeof(new_marker));

			for (int row = 0; row < src_len; ++row)
			{
				memcpy(&temp[row], &module.rows[activeTrack][src_row + row], sizeof(rowStruct));
			}

			if (blockToCopyShift)
			{
				for (int row = get_song_last_row() - 1 - src_len; row >= curRow; --row)
				{
					memcpy(&module.rows[activeTrack][row + src_len], &module.rows[activeTrack][row], sizeof(rowStruct));
				}
			}

			for (int row = 0; row < src_len; ++row)
			{
				memcpy(&module.rows[activeTrack][curRow + row], &temp[row], sizeof(rowStruct));
			}

			free(temp);

			strncpy(module.markerName[curRow], new_marker, sizeof(module.markerName[0]));
			module.marker[curRow] = 1;
			if (curRow + src_len < MAX_ROWS) module.marker[curRow + src_len] = 1;
		}
	}
}



void block_copy_by_name(bool shift)
{
	altMenu = false;
	curShiftHold = false;
	blockToCopyShift = shift;
	blockToCopyName[0] = 0;

	set_input_string(PATTERN_XPOS + 9 + get_pattern_width(), get_cursor_row_y(), MAX_MARKER_NAME + 1, MODE_EDIT_SONG, true, blockToCopyName, block_copy_by_name_cb);
}



void block_find_parent(void)
{
	int row = block_find_beginning(curRow);

	if (module.marker[row])
	{
		if (strlen(module.markerName[row]) > 0)
		{
			if(module.markerName[row][0]=='@')
			{
				int parent_row = block_find_by_name(&module.markerName[row][1]);
				
				if (parent_row >= 0) curRow = parent_row;
			}
		}
	}
}



void blocks_update_by_references(void)
{
	set_undo();

	for (int row = 0; row < MAX_ROWS; ++row)
	{
		if (module.marker[row])
		{
			if (strlen(module.markerName[row]) > 0)
			{
				if (module.markerName[row][0] == '@')
				{
					int parent_row = block_find_by_name(&module.markerName[row][1]);

					if (parent_row >= 0)
					{
						int parent_len = block_get_length(parent_row);

						for (int i = 0; i < parent_len; ++i)
						{
							memcpy(&module.rows[activeTrack][row + i], &module.rows[activeTrack][parent_row + i], sizeof(rowStruct));
						}
					}
				}
			}
		}
	}
}



void editor_events(SDL_Event event)
{
	bool shift=event.key.keysym.mod&(KMOD_LSHIFT|KMOD_RSHIFT)?true:false;
	bool ctrl= event.key.keysym.mod&(KMOD_LCTRL |KMOD_RCTRL )?true:false;

	int key = event.key.keysym.sym;

	if (event.type == SDL_KEYDOWN)
	{
		if (key == SDLK_LCTRL || key == SDLK_RCTRL) altMenu = true;
	}
	if (event.type == SDL_KEYUP)
	{
		if (key == SDLK_LCTRL || key == SDLK_RCTRL) altMenu = false;
	}

	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			int ct=module.columnType[curColumn];

			if (!altMenu)
			{
				switch (key)
				{
					case SDLK_F1:
					{
						set_edit_mode(MODE_HELP);
					}
					break;

					case SDLK_F2:
					{
						set_select_file("Save song file", songDirectory, sizeof(songDirectory), MODE_EDIT_SONG, SONG_EXTENSION, true, EVENT_SAVE_SONG, selectFileSongLast, sizeof(selectFileSongLast), 1, 0);
					}
					break;

					case SDLK_F3:
					{
						set_select_file("Load song file", songDirectory, sizeof(songDirectory), MODE_EDIT_SONG, SONG_EXTENSION, true, EVENT_LOAD_SONG, selectFileSongLast, sizeof(selectFileSongLast), 0, 0);
					}
					break;

					case SDLK_F4:
					{
						set_instrument_editor();
					}
					break;

					case SDLK_F5:
					{
						if(!shift) music_toggle(false);
					}
					break;

					case SDLK_F6:
					{
						set_info();
					}
					break;

					case SDLK_F7:
					{
						set_select_engine();
					}
					break;

					case SDLK_F8:
					{
						set_export_data();
					}
					break;
				}
			}
			else
			{
				switch (key)
				{
					case SDLK_F1:
					{
						if (editStyle == EDIT_STYLE_CHOPPED) editStyle = EDIT_STYLE_LINEAR; else editStyle = EDIT_STYLE_CHOPPED;
					}
					break;

					case SDLK_F2:
					{
						if (!quietMode) quietMode = true; else quietMode = false;
					}
					break;

					case SDLK_F3:
					{
						do_backup();
						module_clear();
					}
					break;

					case SDLK_F4:
					{
						module_optimize_row_speed();
					}
					break;

					case SDLK_F5:
					{
						dualMode ^= true;
					}
					break;

					case SDLK_F6:
					{
						activeTrack ^= 1;
					}
					break;

					case SDLK_F8:
					{
						set_export_audio();
					}
					break;
				}
			}

			if(!ctrl)
			{
				switch(key)
				{
				case SDLK_UP://move cursor up one row
					{
						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_DOWN://move cursor down one row
					{
						move_row_cursor(1);
						update_selection();
					}
					break;

				case SDLK_PAGEUP://move cursor page up
					{
						move_row_cursor(-curPageStep);
						update_selection();
					}
					break;

				case SDLK_PAGEDOWN://move cursor page down
					{
						move_row_cursor(curPageStep);
						update_selection();
					}
					break;

				case SDLK_HOME://move cursor to a previous marker
					{
						move_row_cursor_prev_marker();
						update_selection();
					}
					break;

				case SDLK_END://move cursor to a next marker
					{
						move_row_cursor_next_marker();
						update_selection();
					}
					break;

				case SDLK_LEFT://move cursor left
					{
						--curColumn;

						if (curColumn < 0) curColumn = module.columns - 1;

						while (module.columnType[curColumn] == COLUMN_TYPE_EMPTY) --curColumn;

						update_selection();
					}
					break;

				case SDLK_RIGHT://move cursor right
					{
						++curColumn;

						if (curColumn >= module.columns) curColumn = 0;

						while (module.columnType[curColumn] == COLUMN_TYPE_EMPTY) ++curColumn;

						update_selection();
					}
					break;

				case SDLK_TAB://move cursor to the next or previous channel
					{
						if(shift) change_channel(-1); else change_channel(1);

						update_selection();
					}
					break;

				case SDLK_DELETE://clear a field under cursor or a selection
					{
						if(curSelectionWidth&&curSelectionHeight)
						{
							selection_copy_cut(false,true,false);
						}
						else
						{
							set_undo();

							module.rows[activeTrack][curRow].column[curColumn]=0;
							move_row_cursor(curAutoStep);
							update_selection();
						}
					}
					break;

				case SDLK_BACKQUOTE://increment or decrement autostep
					{
						if(shift)
						{
							if(curAutoStep) --curAutoStep;
						}
						else
						{
							if(curAutoStep<99) ++curAutoStep;
						}
					}
					break;

				case SDLK_KP_MULTIPLY://increment octave
					{
						if(curOctave<9) ++curOctave;
					}
					break;

				case SDLK_KP_MINUS://decrement octave
					{
						--module.speed;

						if (module.speed < module.speed_min) module.speed = module.speed_min;
					}
					break;

				case SDLK_KP_PLUS://increase global speed
					{
						++module.speed;

						if (module.speed > module.speed_max) module.speed = module.speed_max;
					}
					break;

				case SDLK_KP_DIVIDE://decrease global speed
					{
						if(curOctave) --curOctave;
					}
					break;

				case SDLK_LSHIFT://start selection
				case SDLK_RSHIFT:
					{
						curShiftHold=true;

						if(!curSelectionActive)
						{
							curSelectionXS=curColumn;
							curSelectionYS=curRow;
							curSelectionActive=true;
						}

						update_selection();
					}
					break;

				case SDLK_SPACE://toggle a marker
					{
						module.marker[curRow]^=1;
					}
					break;

				case SDLK_INSERT://insert an empty position into current channel with shift
					{
						set_undo();

						int channel = module.columnChannel[curColumn];

						for (int col = 0; col < MAX_COLUMNS; ++col)
						{
							if (module.columnChannel[col] == channel) insert_pos_column(curRow, col);
						}
					}
					break;

				case SDLK_BACKSPACE://delete position of current channel with shift
					{
						set_undo();

						int channel = module.columnChannel[curColumn];

						for (int col = 0; col < MAX_COLUMNS; ++col)
						{
							if (module.columnChannel[col] == channel) delete_pos_column(curRow, col);
						}

						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_QUOTE:	//repeat a previously entered value
					{
						if (curPreviousValue >= 0)
						{
							if (curPreviousValueType == ct)
							{
								set_undo();

								module.rows[activeTrack][curRow].column[curColumn] = curPreviousValue;

								if (!quietMode) music_play_row();

								move_row_cursor(curAutoStep);
							}
						}
					}
					break;

				case SDLK_8:
					{
						if(!shift) music_play_row();
					}
					break;

				case SDLK_RETURN:
					{
						if(!curEnterHold)
						{
							curEnterHold=true;
							music_play_from_cursor();
						}
					}
					break;
				}

				if(!ctrl&&!shift)
				{
					switch(ct)
					{
					case COLUMN_TYPE_NOTE://enter note
						{
						for (int i = 0; i < noteKeysCount; i += 2)
							{
								if(noteKeys[i]==key)
								{
									set_undo();

									int note = noteKeys[i + 1];

									if (note < 0)
									{
										note = 1;
									}
									else
									{
										note = 2 + note + curOctave * 12;

										if (note < module.lowestNote)  note = module.lowestNote;
										if (note > module.highestNote) note = module.highestNote;
									}

									module.rows[activeTrack][curRow].column[curColumn] = note;

									curPreviousValue = note;
									curPreviousValueType = ct;

									if (note != 1 && module.columnPlay[curColumn]) if (!quietMode) music_play_row();

									move_row_cursor(curAutoStep);

									break;
								}
							}
						}
						break;

					case COLUMN_TYPE_LETTER:	//enter a letter A..Z
						{
							if (key >= SDLK_a && key <= SDLK_z)
							{
								set_undo();

								int letter = key - SDLK_a + 'A';

								module.rows[activeTrack][curRow].column[curColumn] = letter;

								curPreviousValue = letter;
								curPreviousValueType = ct;

								if (module.columnPlay[curColumn]) if (!quietMode) music_play_row();

								move_row_cursor(curAutoStep);
							}
						}
						break;

					default://enter a numeric value
						{
							if(key>=SDLK_0&&key<=SDLK_9)
							{
								set_undo();

								int value = 0;

								if(curResetValue)
								{
									curResetValue=false;
								}
								else
								{
									value = module.rows[activeTrack][curRow].column[curColumn];
								}

								value = (value * 10 + (key - SDLK_0)) % (int)pow(10.0f, ct);

								if (value > module.columnMaxValue[curColumn]) value = module.columnMaxValue[curColumn];

								module.rows[activeTrack][curRow].column[curColumn] = value;

								curPreviousValue = value;
								curPreviousValueType = ct;

								if (value&&module.columnPlay[curColumn]) if (!quietMode) music_play_row();

								++curAutoStepCount;

								if (curAutoStepCount >= module.columnType[curColumn])
								{
									move_row_cursor(curAutoStep);
								}
							}
						}
					}

					int octave = -1;//select octave

					switch (key)
					{
					case SDLK_KP_0: octave = 0; break;
					case SDLK_KP_1: octave = 1; break;
					case SDLK_KP_2: octave = 2; break;
					case SDLK_KP_3: octave = 3; break;
					case SDLK_KP_4: octave = 4; break;
					case SDLK_KP_5: octave = 5; break;
					case SDLK_KP_6: octave = 6; break;
					case SDLK_KP_7: octave = 7; break;
					case SDLK_KP_8: octave = 8; break;
					case SDLK_KP_9: octave = 9; break;
					}
					if (octave >= 0) curOctave = octave;
				}
			}

			if(ctrl)
			{
				switch(key)
				{
				case SDLK_PAGEUP:////move cursor large page up
					{
						move_row_cursor(-curPageLargeStep);
						update_selection();
					}
					break;

				case SDLK_PAGEDOWN://move cursor large page down
					{
						move_row_cursor(curPageLargeStep);
						update_selection();
					}
					break;

				case SDLK_n:
				case SDLK_HOME://set loop start
					{
						if(module.loopSupport)
						{
							set_undo();
							module.loopStart=curRow;
						}
					}
					break;

				case SDLK_m:
				case SDLK_END://set loop end
					{
						if(module.loopSupport)
						{
							set_undo();
							module.loopEnd=curRow;
						}
					}
					break;

				case SDLK_a://select all channels
					{
						select_channel(-1);
					}
					break;

				case SDLK_l://select current channel
					{
						select_channel(curColumn);
					}
					break;

				case SDLK_f://name a block
				{
					block_set_name();
				}
				break;

				case SDLK_q://copy a block by name
				{
					block_copy_by_name(false);
				}
				break;

				case SDLK_p://find parent to current reference block, or update all referenced blocks
				{
					if (!shift)
					{
						block_find_parent();
					}
					else
					{
						blocks_update_by_references();
					}
				}
				break;

				case SDLK_g://go to line
					{
						altMenu = false; 
					
						set_input_number(PATTERN_XPOS + 2, get_cursor_row_y(), 4, MODE_EDIT_SONG, true, &curRow, false, NULL);
					}
					break;

				case SDLK_z://undo
					{
						undo();
					}
					break;

				case SDLK_x://cut selection
					{
						selection_copy_cut(true,true,false);
					}
					break;

				case SDLK_c://copy selection
					{
						selection_copy_cut(true,false,false);
					}
					break;

				case SDLK_v://paste selection
					{
						selection_paste(false);
					}
					break;

				case SDLK_0://unmute all
					{
						unmute_all();					
					}
					break;

				case SDLK_BACKQUOTE://mute all but current channel
					{
						mute_all_but_current();
					}
					break;

				case SDLK_INSERT://insert an empty row with shift
					{
						set_undo();
						insert_row_with_shift(curRow);
					}
					break;

				case SDLK_BACKSPACE://delete a row with shift
					{
						set_undo();
						delete_row_with_shift(curRow);
						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_e://expand selection
					{
						selection_expand();
					}
					break;

				case SDLK_s://shrink selection
					{
						selection_shrink();
					}
					break;

				case SDLK_u://transpose semitone up
					{
						selection_transpose(1);
					}
					break;

				case SDLK_d://transpose semitone down
					{
						selection_transpose(-1);
					}
					break;

				case SDLK_o://transpose octave up
					{
						selection_transpose(12);
					}
					break;

				case SDLK_k://transpose octave up
					{
						selection_transpose(-12);
					}
					break;

				case SDLK_r:
					{
						selection_replicate();
					}
					break;

				case SDLK_b:
					{
						if(!shift) selection_borrow_tracks(true);
					}
					break;
					
				case SDLK_t:
					{
						selection_exchange_tracks();
					}
					break;

				case SDLK_w:
					{
						selection_unwind_notes();
					}
					break;

				case SDLK_LEFTBRACKET://decrement measure
					{
						if(module.measure>2) --module.measure;
					}
					break;

				case SDLK_RIGHTBRACKET://increment measure
					{
						if(module.measure<1024) ++module.measure;
					}
					break;

				case SDLK_EQUALS://increase global speed
					{
						if(module.speed<99) ++module.speed;
					}
					break;

				case SDLK_MINUS://decrease global speed
					{
						if(module.speed>1) --module.speed;
					}
					break;
				}

				if (key >= SDLK_1  &&key <= SDLK_9) channelMute[activeTrack][key - SDLK_1] ^= true;//toggle mute for channel 1-9
				
				int chn = -1;
				switch (key)
				{
				case SDLK_KP_1: chn = 0; break;
				case SDLK_KP_2: chn = 1; break;
				case SDLK_KP_3: chn = 2; break;
				case SDLK_KP_4: chn = 3; break;
				case SDLK_KP_5: chn = 4; break;
				case SDLK_KP_6: chn = 5; break;
				case SDLK_KP_7: chn = 6; break;
				case SDLK_KP_8: chn = 7; break;
				case SDLK_KP_9: chn = 8; break;
				}
				if (chn >= 0) channelMute[activeTrack][chn] ^= true;
			}

			if(shift&&!ctrl)
			{
				switch(key)
				{
				case SDLK_F5://play from row
					{
						if (!altMenu) music_toggle(true);
					}
					break;

				case SDLK_x://cut selection with shift
					{
						selection_copy_cut(true,true,true);
					}
					break;

				case SDLK_c://copy selection
					{
						selection_copy_cut(true,false,false);
					}
					break;

				case SDLK_v://paste selection with shift
					{
						selection_paste(true);
					}
					break;

				case SDLK_q://copy a block by name with shift
				{
					block_copy_by_name(true);
				}
				break;

				case SDLK_b:
					{
						selection_borrow_tracks(false);
					}
					break;
				}

				if(key>=SDLK_0&&key<=SDLK_9)//select octave
				{
					curOctave=key-SDLK_0;
				}
			}
		}
		break;

	case SDL_KEYUP:
		{
			switch(key)
			{
			case SDLK_LSHIFT:
			case SDLK_RSHIFT:
				{
					curShiftHold=false;
				}
				break;

			case SDLK_RETURN:
				{
					if(curEnterHold)
					{
						curEnterHold=false;
						music_stop();
					}
				}
				break;
			}

			/*for(i=0;i<noteKeysCount;i+=2)
			{
				if(noteKeys[i]==key)
				{
					music_stop();
					break;
				}
			}*/
		}
		break;
	}
}